# Module globals
adjustmentSpecFileNames <- GetAdjustmentSpecFileNames()

# User interface
dataAdjustUI <- function(id)
{
  ns <- NS(id)

  tagList(
    box(
      width = 12,
      title = "Run adjustments",
      solidHeader = TRUE,
      status = "primary",
      actionButton(ns("addAdjustBtn"), "Add"),
      p(""),
      div(id = ns("adjustmentList")),
      actionButton(ns("runAdjustBtn"), "Run")),
    box(
      width = 12,
      title = "Run log",
      solidHeader = TRUE,
      status = "warning",
      verbatimTextOutput(ns("runLog"))),
    box(
      width = 12,
      title = "Adjusted data details",
      solidHeader = TRUE,
      status = "warning",
      uiOutput(ns("dataDetails")))
  )
}

# Server logic
dataAdjust <- function(input, output, session, inputData)
{
  # Get namespace
  ns <- session$ns

  # Store reactive values
  finalData <- reactiveVal(NULL)
  adjustedData <- reactiveVal(NULL)
  vals <- reactiveValues(runLog = "",
                         adjustmentSpecs = list(),
                         #adjustmentSelectionWidgets = tagList(),
                         lastAdjustmentWidgetIndex = 0L,
                         editedAdjustmentKey = "",
                         editedAdjustmentParamsWidgets = list())

  # Reactive values observers
  observeEvent(vals$adjustmentSpecs, {
    finalData(NULL)
    adjustedData(NULL)
  })

  # Add adjustment selection widget
  observeEvent(input[["addAdjustBtn"]], {
    widget <- GetAdjustmentSelectionWidget()
    insertUI(selector = paste0("#", ns("adjustmentList")),
             where = "beforeEnd",
             ui = widget)
  })

  # EVENT: Button "Download intermediate data" clicked
  DownloadIntermediateData <- function(key) {
    adjustedData <- adjustedData()
    if (key %in% names(adjustedData)) {
      intermediateData <- adjustedData[[key]]$Table
    } else {
      intermediateData <- NULL
    }

    downloadHandler(
      filename = function() {
        fileName <- paste("data", "csv", sep = ".")
        return(fileName)
      },
      content = function(file) {
        withProgress(message = "Creating intermediate data output file",
                     detail = "The file will be available for download shortly.",
                     value = 0, {
                       setProgress(0)
                       WriteDataFile(intermediateData, file)
                       setProgress(1)
                     })
        return(NULL)
      })
  }

  # Get widget for selecting an adjustment
  GetAdjustmentSelectionWidget <- function() {
    # Get unique index for the elements
    index <- vals$lastAdjustmentWidgetIndex + 1
    key <- as.character(index)

    rowId          <- paste0("row", key)
    deleteBtnId    <- paste0("deleteBtn", key)
    adjustSelectId <- paste0("adjustSelect", key)
    editParamBtnId <- paste0("adjustParamBtn", key)
    downloadBtnId  <- paste0("downloadBtn", key)

    # Get widget html
    widget <- fluidRow(
      id = ns(rowId),
      column(1, actionLink(ns(deleteBtnId), "Delete")),
      # Selection input
      column(5, selectInput(ns(adjustSelectId),
                            label = NULL,
                            choices = adjustmentSpecFileNames)),
      # Edit parameters button
      column(3, actionLink(ns(editParamBtnId), "Edit parameters")),
      column(3, shinyjs::hidden(downloadLink(ns(downloadBtnId), "Download intermediate data"))))

    # EVENT: Adjustment selection changed
    observeEvent(input[[adjustSelectId]], {
      # Add selected adjustment to the list of adjustments to run
      selectedAdjustmentFileName <- input[[adjustSelectId]]
      if (file.exists(selectedAdjustmentFileName)) {
        # Get adjustment specification and enrich with key for later reference
        adjustmentSpec <- GetAdjustmentSpec(selectedAdjustmentFileName)
        adjustmentSpec$Key <- key
        vals$adjustmentSpecs[[key]] <- adjustmentSpec
      }
    })

    # EVENT: Button "Delete" clicked
    observeEvent(input[[deleteBtnId]], {
      vals$adjustmentSpecs[[key]] <- NULL
      removeUI(selector = paste0("#", ns(rowId)))
    })

    # EVENT: Button "Edit parameters" clicked
    observeEvent(input[[editParamBtnId]], {
      vals$editedAdjustmentKey <- key
    })

    # Store for next adjustment selection widget addition
    vals$lastAdjustmentWidgetIndex <- index

    return(widget)
  }

  EnableIntermediateDataDownloadLinks <- function() {
    for (adjustmentSpec in vals$adjustmentSpecs) {
      key <- adjustmentSpec$Key
      downloadBtnId  <- paste0("downloadBtn", key)

      # Define download handler
      output[[downloadBtnId]] <- DownloadIntermediateData(key)
      # Show download link
      shinyjs::show(downloadBtnId)
    }
  }

  # Populate adjustment parameter widgets in the editing dialog
  output[["adjustmentList"]] <- renderUI({
    vals$adjustmentSelectionWidgets
  })

  # Get widgets for editing parameters in a dialog
  GetAdjustParamsWidgets <- function(adjustmentSpec) {
    # Reformat slightly - add "name" property for "GetParamWidgets"
    paramSpecs <- adjustmentSpec$Parameter
    paramSpecs <- setNames(lapply(names(paramSpecs), function(name) {
      paramSpec <- paramSpecs[[name]]
      paramSpec$name <- name
      structure(paramSpec, class = "knit_param")
    }), names(paramSpecs))

    widgets <- GetParamWidgets(paramSpecs, ns = ns)

    return(widgets)
  }

  # Show adjustment parameters editing dialog
  observeEvent(vals$editedAdjustmentKey, {
    if (vals$editedAdjustmentKey != "") {
      editedAdjustmentSpec <- vals$adjustmentSpecs[[vals$editedAdjustmentKey]]
      vals$editedAdjustmentParamsWidgets <- GetAdjustParamsWidgets(editedAdjustmentSpec)

      showModal(modalDialog(
        title = "Edit adjustment parameters",
        uiOutput(ns("adjustmentParams")),
        easyClose = TRUE,
        footer = tagList(
          actionButton(ns("paramsDlgCancel"), "Cancel"),
          actionButton(ns("paramsDlgOk"), "OK")
        )
      ))
    } else {
      vals$editedAdjustmentParamsWidgets <- list()
    }
  })

  # Populate adjustment parameter widgets in the editing dialog
  output[["adjustmentParams"]] <- renderUI({
    vals$editedAdjustmentParamsWidgets
  })

  # Adjustment parameters editing dialog CLOSE through OK event
  observeEvent(input[["paramsDlgOk"]], {
    # Copy parameters from dialog
    adjustmentParams <- vals$adjustmentSpecs[[vals$editedAdjustmentKey]]$Parameters
    for (paramName in names(vals$editedAdjustmentParamsWidgets)) {
      adjustmentParams[[paramName]]$value <- input[[paramName]]
    }
    # Save parameters
    vals$adjustmentSpecs[[vals$editedAdjustmentKey]]$Parameters <- adjustmentParams

    # Clean up
    vals$editedAdjustmentKey <- ""
    vals$editedAdjustmentParamsWidgets <- list()
    removeModal()
  })

  # Adjustment parameters editing dialog CLOSE through Cancel event
  observeEvent(input[["paramsDlgCancel"]], {
    # Clean up
    vals$editedAdjustmentKey <- ""
    vals$editedAdjustmentParamsWidgets <- list()
    removeModal()
  })

  # EVENT: Button "Run adjustments" clicked
  observeEvent(input[["runAdjustBtn"]], {
    # Run adjustments
    finalData(NULL)
    adjustedData(NULL)
    vals$runLog <- "Running adjustments..."
    tryCatch({
      runLog <- capture.output({
        adjustedData <- RunAdjustments(inputData(),
                                       adjustmentSpecs = vals$adjustmentSpecs)
      })
      vals$runLog <- paste(runLog, collapse = "\n")
    }, error = function(err) {
      cat(err)
      adjustedData <- NULL
    }, finally = {
      adjustedData(adjustedData)
    })
  })

  # Populate adjustment parameter widgets in the editing dialog
  output[["runLog"]] <- renderText({
    vals$runLog
  })

  # Observe change in adjustedData object
  observeEvent(adjustedData(), {
    adjustedData <- adjustedData()
    if (length(adjustedData) > 0) {
      finalData(adjustedData[[length(adjustedData)]])
      EnableIntermediateDataDownloadLinks()
    } else {
      finalData(NULL)
    }
  })

  # Output adjusted data details when they have changed
  output[["dataDetails"]] <- renderUI({
    # Respond to adjustedData change
    finalData <- finalData()$Table

    isolate({
      if (!is.null(finalData)) {
        # Get adjusted data details
        finalDataDetails <-
          div(
            p(strong("Number of records:"), nrow(finalData)),
            p(strong("Column names:"), paste(names(finalData), collapse = ", "))
          )
      } else {
        finalDataDetails <- p("Please, select at least one adjustment.")
      }

      return(finalDataDetails)
    })
  })

  return(finalData)
}
