# User interface
inputDataUploadUI <- function(id)
{
  ns <- NS(id)

  tagList(
    box(
      width = 6,
      title = "Specify input data file",
      solidHeader = TRUE,
      status = "primary",
      fileInput(ns("fileInput"), label = h3("File input"))),
    box(
      width = 6,
      title = "Input data basic details",
      solidHeader = TRUE,
      status = "warning",
      uiOutput(ns("dataDetails"))),
    uiOutput(ns("attributesMapping"))
  )
}

# Server logic
inputDataUpload <- function(input, output, session)
{
  # EVENT: Input data file name changed
  inputData <- eventReactive(input[["fileInput"]], {
    # Get reference to file input
    fileInput <- input$fileInput

    # Validate
    validate(need(fileInput, message = FALSE))

    # Get input data file name and type
    fileType <- tolower(tools::file_ext(fileInput$name))
    fileName <- fileInput$datapath

    # Read input data
    inputData <- ReadDataFile(fileName, fileType)
    return(inputData)
  }, ignoreNULL = TRUE)

  # Output input data details when they have changed
  output[["dataDetails"]] <- renderUI({
    # Respond to inputData change
    inputData <- inputData()

    isolate({
      # Get reference to file input
      fileInput <- input$fileInput

      # Get input data details
      inputDataDetails <-
        div(
          p(strong("Name:"), fileInput$name),
          p(strong("Size:"), FormatObjectSize(fileInput$size)),
          p(strong("Type:"), fileInput$type),
          p(strong("Number of records:"), nrow(inputData)),
          p(strong("Column names:"), paste(colnames(inputData), collapse = ", "))
        )

      return(inputDataDetails)
    })
  })

  # Generate attributes mapping
  output[["attributesMapping"]] <- renderUI({
    # Respond to inputData change
    inputData <- inputData()

    isolate({

      ns <- session$ns

      # Get input data details
      analysisAttrs <- c("DiagnosisYear", "Gender", "Age", "Transmission", "Migration", "AIDS")
      dataAttrs <- colnames(inputData)

      # i <- 1
      descrText <- p("Please, provide mapping between attributes used internally by the tool (column \"Tool\") and data dimensions (column \"Data\").")
      headers <- fluidRow(
        column(3, h3("Tool")),
        column(9, h3("Data"))
      )

      attributesMapping <- lapply(seq_along(analysisAttrs), function(i) {
        analysisAttr <- analysisAttrs[i]

        fluidRow(
          # Selection attribute
          column(3, h4(analysisAttr)),
          # Selection data dimension
          column(9, selectizeInput(ns(paste0("dataAttrs", i)),
                                   label = NULL,
                                   choices = dataAttrs,
                                   options = list(
                                     placeholder = paste("Select", analysisAttr, "attribute"),
                                     onInitialize = I('function() { this.setValue(""); }'))
          )))
      })

      # Wrap in a box
      attributesMapping <- box(
        width = 12,
        title = "Attributes mapping",
        solidHeader = TRUE,
        status = "primary",
        tagList(descrText,
                headers,
                attributesMapping))


      return(attributesMapping)
    })
  })

  return(inputData)
}
